import datetime
import os
import tempfile
import time
from typing import Optional

import pexpect
from pytest_embedded.log import DuplicateStdout, DuplicateStdoutPopen
from pytest_embedded.utils import to_str
from pytest_embedded_qemu import DEFAULT_IMAGE_FN


class Qemu(DuplicateStdoutPopen):
    """
    :ivar: qemu_inst: ``subprocess.Popen`` process which runs QEMU program
    :ivar: log_file: serial log filepath
    """

    QEMU_PROG_PATH = 'qemu-system-xtensa'
    QEMU_DEFAULT_ARGS = '-nographic -no-reboot -machine esp32'

    QEMU_STRAP_MODE_FMT = '-global driver=esp32.gpio,property=strap_mode,value={}'
    QEMU_SERIAL_TCP_FMT = '-serial tcp::{},server,nowait'

    def __init__(
        self,
        qemu_image_path: Optional[str] = None,
        qemu_prog_path: Optional[str] = None,
        qemu_cli_args: Optional[str] = None,
        qemu_extra_args: Optional[str] = None,
        qemu_log_path: Optional[str] = None,
        **kwargs,
    ):
        image_path = qemu_image_path or DEFAULT_IMAGE_FN
        if not os.path.exists(image_path):
            raise ValueError(f'QEMU image path not exists: {image_path}')

        qemu_prog_path = qemu_prog_path or self.QEMU_PROG_PATH
        qemu_cli_args = qemu_cli_args or self.QEMU_DEFAULT_ARGS

        qemu_extra_args = qemu_extra_args.replace('"', '') if qemu_extra_args else qemu_extra_args
        qemu_extra_args = [qemu_extra_args] if qemu_extra_args else []
        qemu_extra_args.append(f'-drive file={image_path},if=mtd,format=raw')

        # we use log file to record serial output
        self.log_file = qemu_log_path or os.path.join(
            tempfile.tempdir, datetime.datetime.utcnow().strftime('%Y-%m-%d_%H-%M-%S'), 'serial.log'
        )
        os.makedirs(os.path.dirname(self.log_file), exist_ok=True)
        qemu_extra_args.append(f'-serial file:{self.log_file}')

        self.qemu_inst = None

        cmd = f'{qemu_prog_path} {qemu_cli_args} {" ".join(qemu_extra_args)}'
        super().__init__(cmd, **kwargs)

    def _forward_io(self, pexpect_proc: Optional[pexpect.spawn] = None, source: Optional[str] = None) -> None:
        time.sleep(1)  # ensure the log file is generated by qemu
        with DuplicateStdout(pexpect_proc, source):
            for line in open(self.log_file):
                print(to_str(line))



import logging
import os
from types import ModuleType

import pytest
from pytest_embedded_qemu import DEFAULT_IMAGE_FN

from .dut import QemuDut
from .qemu import Qemu


def pytest_addoption(parser):
    group = parser.getgroup('embedded')
    group.addoption(
        '--qemu-image-path',
        help='QEMU image path. Could be overridden by pytest parametrizing. ' '(Default: "<app_path>/flash_image.bin)',
    )
    group.addoption(
        '--qemu-prog-path',
        help='QEMU program path. (Default: "qemu-system-xtensa")',
    )
    group.addoption(
        '--qemu-cli-args',
        help='QEMU cli default arguments. Could be overridden by pytest parametrizing. '
        '(Default: "-nographic -no-reboot -machine esp32")',
    )
    group.addoption(
        '--qemu-extra-args',
        help='QEMU cli extra arguments, will append to the argument list. '
        'Could be overridden by pytest parametrizing. (Default: None)',
    )
    group.addoption(
        '--qemu-log-path',
        help='QEMU log file path. (Default: "<temp folder>/<timestamp>/serial.log")',
    )


@pytest.fixture
def qemu_image_path(request):
    return {'qemu_image_path': getattr(request, 'param', None)}


@pytest.fixture
def qemu_cli_args(request):
    """
    Apply parametrization to fixture :func:`pytest_embedded_qemu_idf.plugin.qemu`
    """
    return {'qemu_cli_args': getattr(request, 'param', None)}


@pytest.fixture
def qemu_extra_args(request):
    """
    Apply parametrization to fixture :func:`pytest_embedded_qemu_idf.plugin.qemu`
    """
    return {'qemu_extra_args': getattr(request, 'param', None)}


@pytest.hookimpl
def pytest_plugin_registered(plugin, manager):
    if not isinstance(plugin, ModuleType) or plugin.__name__ != 'pytest_embedded.plugin':
        return

    plugin.KNOWN_OPTIONS['Qemu'].extend(
        [
            'qemu_image_path',
            'qemu_prog_path',
            'qemu_cli_args',
            'qemu_extra_args',
            'qemu_log_path',
        ]
    )
    plugin.ENV['qemu'] = True

    if 'idf' in plugin.ENV:
        from .app import QemuApp

        plugin.KNOWN_OPTIONS['App'].append('qemu_image_path')

        @pytest.fixture
        def app(qemu_image_path, pexpect_proc, options, test_file_path) -> QemuApp:
            """
            Uses :attr:`options['App']` as kwargs to create instance.

            :return: :class:`pytest_embedded.app.App` or derived class instance
            """
            app_options = options.get('App', {})
            if app_options['app_path'] is None:
                logging.info(f'test_file_path: {test_file_path}')
                app_path = os.path.dirname(test_file_path)
            else:
                app_path = app_options['app_path']
            app_options['app_path'] = app_path

            if qemu_image_path['qemu_image_path']:
                app_options.update(qemu_image_path)

            if app_options['qemu_image_path'] is None:
                app_options['qemu_image_path'] = os.path.join(app_path, DEFAULT_IMAGE_FN)

            app_options['pexpect_proc'] = pexpect_proc
            return QemuApp(**app_options)

        globals()['app'] = app

    @pytest.fixture
    def qemu(app, options, qemu_image_path, qemu_cli_args, qemu_extra_args) -> Qemu:
        """
        Uses :attr:`options['Qemu']` as kwargs to create instance.

        :return: :class:`pytest_embedded_qemu_idf.qemu.IdfQemu` or derived class instance
        """
        qemu_options = options.get('Qemu', {})
        if qemu_image_path['qemu_image_path']:
            qemu_options.update(qemu_image_path)
        if qemu_cli_args['qemu_cli_args']:
            qemu_options.update(qemu_cli_args)
        if qemu_extra_args['qemu_extra_args']:
            qemu_options.update(qemu_extra_args)

        if qemu_options['qemu_image_path'] is None:
            qemu_options['qemu_image_path'] = os.path.join(app.app_path, DEFAULT_IMAGE_FN)

        qemu = Qemu(**qemu_options)
        try:
            yield qemu
        finally:
            qemu.close()

    globals()['qemu'] = qemu

    @pytest.fixture
    def dut(qemu, app, pexpect_proc, options) -> QemuDut:
        """
        Uses :attr:`options['Dut']` as kwargs to create instance.

        :return: :class:`pytest_embedded_qemu_idf.dut.QemuDut` or derived class instance
        """
        dut_options = options.get('Dut', {})
        dut = QemuDut(qemu, app, pexpect_proc, **dut_options)
        try:
            yield dut
        finally:
            dut.close()

    globals()['dut'] = dut



import os
from typing import Optional

import pexpect
from pytest_embedded.log import cls_redirect_stdout, live_print_call
from pytest_embedded_idf.app import IdfApp

from . import DEFAULT_IMAGE_FN


class IdfFlashImageMaker:
    """
    Create a single image for qemu based on the :class:`pytest_embedded_idf.app.IdfApp`'s partition
    table and all the flash files.

    :ivar: app: :class:`pytest_embedded_idf.app.IdfApp` instance
    :ivar: image_path: output image path, would create with :meth:`make_bin` if not set.
    """

    def __init__(self, app: IdfApp, image_path: str):
        self.app = app
        self.image_path = image_path

    def make_bin(self) -> None:
        """
        Create a single image file for qemu.
        """
        # flash files is sorted, if the first offset is not 0x0, we need to fill in with empty bin
        if self.app.flash_files[0][0] != 0x0:
            self._write_empty_bin(count=self.app.flash_files[0][0])
        for offset, file_path, encrypted in self.app.flash_files:
            if encrypted:
                raise NotImplementedError('will implement later')
            else:
                self._write_bin(file_path, seek=offset)

    def _write_empty_bin(self, count: int, bs: int = 1024, seek: int = 0):
        live_print_call(
            f'dd if=/dev/zero bs={bs} count={count} seek={seek} of={self.image_path}',
            shell=True,
        )

    def _write_bin(self, binary_filepath, bs: int = 1, seek: int = 0):
        live_print_call(
            f'dd if={binary_filepath} bs={bs} seek={seek} of={self.image_path} conv=notrunc',
            shell=True,
        )

    def _write_encrypted_bin(self, binary_filepath, bs: int = 1, seek: int = 0):
        live_print_call(
            f'dd if=/dev/zero bs=1 count=32 of=key.bin',
            shell=True,
        )  # generate a fake key bin
        live_print_call(
            f'espsecure.py encrypt_flash_data --keyfile key.bin --output decrypted.bin --address {seek} '
            f'{binary_filepath}',
            shell=True,
        )
        self._write_bin('decrypted.bin', bs=bs, seek=seek)

    def _burn_efuse(self):
        pass


class QemuApp(IdfApp):
    def __init__(
        self,
        app_path: Optional[str] = None,
        part_tool: Optional[str] = None,
        qemu_image_path: Optional[str] = None,
        pexpect_proc: Optional[pexpect.spawn] = None,
        **kwargs,
    ):
        self.pexpect_proc = pexpect_proc
        self.image_path = qemu_image_path or os.path.join(app_path, DEFAULT_IMAGE_FN)

        super().__init__(app_path, part_tool, **kwargs)

        if self.target != 'esp32':
            raise ValueError('For now on QEMU we only support ESP32')

        self.create_image()

    @cls_redirect_stdout(source='create image')
    def create_image(self) -> None:
        """
        Create the image if not exists
        """
        if os.path.exists(self.image_path):
            print(f'Using image already exists: {self.image_path}')
        else:
            image_maker = IdfFlashImageMaker(self, self.image_path)
            image_maker.make_bin()
